@@ -31,12 +31,10 @@
     int resultCode = _func(handle, geometry1, geometry2, &pResult[i]);        \
                                                                   \
     if (resultCode == 0) {                                        \
-      UNPROTECT(1);                                               \
-      GEOS_ERROR("[i=%d] ", i + 1);                               \
+      Rf_error("[%d] %s", i + 1, globalErrorMessage);             \
     }                                                             \
   }                                                               \
                                                                   \
-  GEOS_FINISH();                                                  \
   UNPROTECT(1);                                                   \
   return result;
 
@@ -57,6 +55,55 @@ SEXP geos_c_distance_frechet(SEXP geom1, SEXP geom2) {
   GEOS_DIST(GEOSFrechetDistance_r);
 }
 
+SEXP geos_c_prepared_distance(SEXP geom1, SEXP geom2) {
+#if LIBGEOS_VERSION_COMPILE_INT >= LIBGEOS_VERSION_INT(3, 9, 1)
+  if (libgeos_version_int() < LIBGEOS_VERSION_INT(3, 9, 1)) {
+    ERROR_OLD_LIBGEOS("GEOSPreparedDistance_r()", "3.9.1");
+  }
+
+  R_xlen_t size = Rf_xlength(geom1);
+  SEXP result = PROTECT(Rf_allocVector(REALSXP, size));
+  double* pResult = REAL(result);
+
+  GEOS_INIT();
+
+  SEXP item1;
+  SEXP item2;
+  GEOSGeometry* geometry1;
+  GEOSGeometry* geometry2;
+  for (R_xlen_t i = 0; i < size; i++) {
+    item1 = VECTOR_ELT(geom1, i);
+    item2 = VECTOR_ELT(geom2, i);
+
+    if (item1 == R_NilValue || item2 == R_NilValue) {
+      pResult[i] = NA_REAL;
+      continue;
+    }
+
+    geometry1 = (GEOSGeometry*) R_ExternalPtrAddr(item1);
+    GEOS_CHECK_GEOMETRY(geometry1, i);
+    geometry2 = (GEOSGeometry*) R_ExternalPtrAddr(item2);
+    GEOS_CHECK_GEOMETRY(geometry2, i);
+
+    const GEOSPreparedGeometry* prepared1 = geos_common_geometry_prepared(item1);
+    if (prepared1 == NULL) {
+      Rf_error("[%d] %s", i + 1, globalErrorMessage); // # nocov
+    }
+
+    int resultCode = GEOSPreparedDistance_r(handle, prepared1, geometry2, &pResult[i]);
+
+    if (resultCode == 0) {
+      Rf_error("[%d] %s", i + 1, globalErrorMessage);
+    }
+  }
+
+  UNPROTECT(1);
+  return result;
+#else
+  ERROR_OLD_LIBGEOS_BUILD("GEOSPreparedDistance_r()", "3.9.1");
+#endif
+}
+
 #define GEOS_DIST_DENSIFY(_func)                                \
   R_xlen_t size = Rf_xlength(geom1);                                          \
   SEXP result = PROTECT(Rf_allocVector(REALSXP, size));                       \
@@ -86,13 +133,11 @@ SEXP geos_c_distance_frechet(SEXP geom1, SEXP geom2) {
     int resultCode = _func(handle, geometry1, geometry2, densifyFracDouble, &pResult[i]);  \
                                                                               \
     if (resultCode == 0) {                                          \
-      UNPROTECT(1);                                                           \
-      GEOS_ERROR("[i=%d] ", i + 1);                                           \
+      Rf_error("[%d] %s", i + 1, globalErrorMessage);                                           \
     }                                                                         \
   }                                                                           \
                                                                               \
-  GEOS_FINISH();                                                              \
-  UNPROTECT(1);                                                               \
+    UNPROTECT(1);                                                               \
   return result;
 
 SEXP geos_c_distance_hausdorff_densify(SEXP geom1, SEXP geom2, SEXP densifyFrac) {
@@ -133,15 +178,13 @@ SEXP geos_c_distance_frechet_densify(SEXP geom1, SEXP geom2, SEXP densifyFrac) {
                                                                               \
     itemResult = _func(handle, geometry1, geometry2);                         \
     if (itemResult == -1) {                                                   \
-      UNPROTECT(1);                                                           \
-      GEOS_ERROR("[i=%d] ", i + 1);                                           \
+      Rf_error("[%d] %s", i + 1, globalErrorMessage);                                           \
     }                                                                         \
                                                                               \
     pResult[i] = itemResult;                                                  \
   }                                                                           \
                                                                               \
-  GEOS_FINISH();                                                              \
-  UNPROTECT(1);                                                               \
+    UNPROTECT(1);                                                               \
   return result;
 
 SEXP geos_c_project(SEXP geom1, SEXP geom2) {
@@ -181,14 +224,12 @@ SEXP geos_c_project_normalized(SEXP geom1, SEXP geom2) {
     int resultCode = _func(handle, geometry1, geometry2);                     \
                                                                               \
     if (resultCode == 2) {                                                    \
-      UNPROTECT(1);                                                           \
-      GEOS_ERROR("[i=%d] ", i + 1);                                           \
+      Rf_error("[%d] %s", i + 1, globalErrorMessage);                                           \
     }                                                                         \
     pResult[i] = resultCode;                                                  \
   }                                                                           \
                                                                               \
-  GEOS_FINISH();                                                              \
-  UNPROTECT(1);                                                               \
+    UNPROTECT(1);                                                               \
   return result;
 
 
@@ -263,15 +304,13 @@ SEXP geos_c_equals_exact(SEXP geom1, SEXP geom2, SEXP tolerance) {
 
     // don't know how to make this fire
     if (resultCode == 2) {
-      UNPROTECT(1); // # nocov
-      GEOS_ERROR("[i=%d] ", i + 1); // # nocov
+      Rf_error("[%d] %s", i + 1, globalErrorMessage); // # nocov
     }
 
     pResult[i] = resultCode;
   }
 
-  GEOS_FINISH();
-  UNPROTECT(1);
+    UNPROTECT(1);
   return result;
 }
 
@@ -301,19 +340,17 @@ for (R_xlen_t i = 0; i < size; i++) {                                       \
   geometry2 = (GEOSGeometry*) R_ExternalPtrAddr(item2);                     \
   GEOS_CHECK_GEOMETRY(geometry2, i);                                        \
                                                                             \
-  const GEOSPreparedGeometry* prepared = GEOSPrepare_r(handle, geometry1);  \
-  GEOS_CHECK_GEOMETRY(prepared, i);                                         \
+  const GEOSPreparedGeometry* prepared = geos_common_geometry_prepared(item1);  \
+  if (prepared == NULL) {                                                   \
+    Rf_error("[%d] %s", i + 1, globalErrorMessage);                         \
+  }                                                                         \
   int resultCode = _func(handle, prepared, geometry2);                      \
-  GEOSPreparedGeom_destroy_r(handle, prepared);                             \
-                                                                            \
   if (resultCode == 2) {                                                    \
-    UNPROTECT(1);                                                           \
-    GEOS_ERROR("[i=%d] ", i + 1);                                           \
+    Rf_error("[%d] %s", i + 1, globalErrorMessage);                         \
   }                                                                         \
   pResult[i] = resultCode;                                                  \
 }                                                                           \
                                                                             \
-GEOS_FINISH();                                                              \
 UNPROTECT(1);                                                               \
 return result;
 
@@ -389,16 +426,14 @@ SEXP geos_c_relate(SEXP geom1, SEXP geom2, SEXP boundaryNodeRule) {
 
     // don't know how to make this fire
     if (itemResult == NULL) {
-      UNPROTECT(1); // # nocov
-      GEOS_ERROR("[i=%d] ", i + 1); // # nocov
+      Rf_error("[%d] %s", i + 1, globalErrorMessage); // # nocov
     }
 
     SET_STRING_ELT(result, i, Rf_mkChar(itemResult));
     GEOSFree_r(handle, itemResult);
   }
 
-  GEOS_FINISH();
-  UNPROTECT(1);
+    UNPROTECT(1);
   return result;
 }
 
@@ -429,7 +464,6 @@ SEXP geos_c_relate_pattern_match(SEXP match, SEXP pattern) {
     pResult[i] = itemResult;
   }
 
-  GEOS_FINISH();
-  UNPROTECT(1);
+    UNPROTECT(1);
   return result;
 }
