@@ -5,26 +5,21 @@
 
 #include <stdio.h>
 
-SEXP geos_c_strtree_create(SEXP geom) {
+SEXP geos_c_strtree_create(SEXP geom, SEXP node_capacity) {
   R_xlen_t size = Rf_xlength(geom);
-
-  // size == 0 will crash some GEOS builds, so we need a sentinel
-  if (size == 0) {
-    SEXP treeIndices = PROTECT(Rf_allocVector(INTSXP, 0));
-    SEXP treeExternalPtr = geos_common_tree_xptr(NULL, geom, treeIndices);
-    UNPROTECT(1);
-    return treeExternalPtr;
-  }
-
+  int nodeCapacityInt = INTEGER(node_capacity)[0];
   GEOS_INIT();
 
   // create the tree object
-  GEOSSTRtree* tree = GEOSSTRtree_create_r(handle, size);
+  GEOSSTRtree* tree = GEOSSTRtree_create_r(handle, nodeCapacityInt);
+  if (tree == NULL) {
+    GEOS_ERROR("%s", "Failed to create GEOSSTRtree*()"); // # nocov
+  }
 
-  // also needed is a sequence of integers to which we can store
+  // also needed is a sequence of indices to which we can store
   // void* pointers, which are the 'hook' for GEOSSTRtree_insert_r()
-  SEXP geomIndices = PROTECT(Rf_allocVector(INTSXP, size));
-  int* pGeomIndices = INTEGER(geomIndices);
+  SEXP geomIndices = PROTECT(Rf_allocVector(REALSXP, size));
+  double* pGeomIndices = REAL(geomIndices);
 
   SEXP item;
   GEOSGeometry* geometry;
@@ -32,8 +27,7 @@ SEXP geos_c_strtree_create(SEXP geom) {
     item = VECTOR_ELT(geom, i);
     if (item == R_NilValue) {
       GEOSSTRtree_destroy_r(handle, tree);
-      GEOS_FINISH();
-      UNPROTECT(1); // geomIndices
+            UNPROTECT(1); // geomIndices
       Rf_error("Can't insert NULL into a geos_str_tree()");
     }
 
@@ -43,22 +37,49 @@ SEXP geos_c_strtree_create(SEXP geom) {
     // store the index of the geometry ("from whence it came")
     // store R 1-based index
     pGeomIndices[i] = i + 1;
-    GEOSSTRtree_insert_r(handle, tree, geometry, &(pGeomIndices[i]));
+    GEOSSTRtree_insert_r(handle, tree, geometry, pGeomIndices + i);
   }
 
-  // tree holds reference to geom to protect it (and the GEOSGeometry* pointers)
-  // from garbage collection
-  SEXP treeExternalPtr = geos_common_tree_xptr(tree, geom, geomIndices);
-  GEOS_FINISH();
-  UNPROTECT(1); // geomIndices
+  // Each call to a query function will need to accumulate indices.
+  // in the absence of std::vector(), we alloc swap space of the
+  // maximum size that would be needed (Rf_xlength(geomIndices)). Another
+  // option would be to implement a growable vector using realloc
+  // beucase it's highly unlikely that any query will actually return
+  // all the elements.
+  SEXP geomIndicesSwap = PROTECT(Rf_allocVector(REALSXP, size));
+
+  // We only get two SEXPs to protect from garbage collection with an externalptr
+  // and we'll used one of to protect `geom`. Use a list() to get around this
+  // for the second one.
+  SEXP externalPtrProt = PROTECT(Rf_allocVector(VECSXP, 2));
+  SET_VECTOR_ELT(externalPtrProt, 0, geomIndices);
+  SET_VECTOR_ELT(externalPtrProt, 1, geomIndicesSwap);
+
+  SEXP treeExternalPtr = geos_common_tree_xptr(tree, geom, externalPtrProt);
+    UNPROTECT(3); // geomIndices, geomIndicesSwap, externalPtrProt
   return treeExternalPtr;
 }
 
-// re-extract the geometries used to construct the tree
+// Re-extract the geometries used to construct the tree
+// not necessary for querying but necessary for distance calculations,
+// and in this form, probably not faster
 SEXP geos_c_strtree_data(SEXP treeExternalPtr) {
   return R_ExternalPtrTag(treeExternalPtr);
 }
 
+// Extract the 1:n vector that holds the index "tag" for each item
+static inline SEXP geos_strtree_indices(SEXP treeExternalPtr) {
+  return VECTOR_ELT(R_ExternalPtrProtected(treeExternalPtr), 0);
+}
+
+// Extract the mutable swap space used to accumulate matching values of
+// geos_strtree_indices() while accumulating indices. Another option would be
+// to implement a growable vector that allocs on the C heap (because we can't
+// longjmp from within a callback).
+static inline SEXP geos_strtree_swap(SEXP treeExternalPtr) {
+  return VECTOR_ELT(R_ExternalPtrProtected(treeExternalPtr), 1);
+}
+
 // data structure and callback used for the generic query
 struct QueryResult {
   GEOSContextHandle_t handle;
@@ -67,7 +88,7 @@ struct QueryResult {
   GEOSPreparedGeometry* prepared;
   SEXP geom;
   SEXP extra;
-  int* indexList;
+  double* indexList;
   R_xlen_t currentIndex;
 };
 
@@ -77,8 +98,7 @@ struct QueryResult {
 SEXP strtree_query_base(SEXP treeExternalPtr, SEXP geom, GEOSQueryCallback callback,
                         int prepare, SEXP extra) {
   GEOSSTRtree* tree = (GEOSSTRtree*) R_ExternalPtrAddr(treeExternalPtr);
-
-  if (tree == NULL && Rf_xlength(R_ExternalPtrTag(treeExternalPtr)) > 0) {
+  if (tree == NULL) {
     UNPROTECT(1);
     Rf_error("External pointer (geos_strtree) is not valid");
   }
@@ -87,12 +107,6 @@ SEXP strtree_query_base(SEXP treeExternalPtr, SEXP geom, GEOSQueryCallback callb
   R_xlen_t size = Rf_xlength(geom);
   SEXP result = PROTECT(Rf_allocVector(VECSXP, size));
 
-  // allocate a temporary integer() along the length of the tree data
-  // that will be used to contain the results while iterating
-  SEXP tempItemResult =  PROTECT(
-    Rf_allocVector(INTSXP, Rf_xlength(R_ExternalPtrProtected(treeExternalPtr)))
-  );
-
   GEOS_INIT();
 
   struct QueryResult queryResult = {
@@ -101,7 +115,7 @@ SEXP strtree_query_base(SEXP treeExternalPtr, SEXP geom, GEOSQueryCallback callb
     .prepared = NULL,
     .geom = geos_c_strtree_data(treeExternalPtr),
     .extra = extra,
-    .indexList = INTEGER(tempItemResult),
+    .indexList = REAL(geos_strtree_swap(treeExternalPtr)),
     .currentIndex = 0
   };
 
@@ -121,8 +135,10 @@ SEXP strtree_query_base(SEXP treeExternalPtr, SEXP geom, GEOSQueryCallback callb
 
     const GEOSPreparedGeometry* prepared;
     if (prepare) {
-      prepared = GEOSPrepare_r(handle, geometry);
-      GEOS_CHECK_GEOMETRY(prepared, i);
+      prepared = geos_common_geometry_prepared(item);
+      if (prepared == NULL) {
+        Rf_error("[%d] %s", i + 1, globalErrorMessage);
+      }
     } else {
       prepared = NULL;
     }
@@ -134,25 +150,19 @@ SEXP strtree_query_base(SEXP treeExternalPtr, SEXP geom, GEOSQueryCallback callb
     queryResult.geometry = geometry;
     queryResult.prepared = (GEOSPreparedGeometry*) prepared;
 
-    // tree is only NULL here as an empty tree
-    if (tree != NULL) {
-      GEOSSTRtree_query_r(handle, tree, geometry, callback, &queryResult);
-    }
-
-    GEOSPreparedGeom_destroy_r(handle, prepared);
+    GEOSSTRtree_query_r(handle, tree, geometry, callback, &queryResult);
 
     // at this point, queryResult now holds the indices of potential intersectors to `geometry`
     // allocate a new vector with the appropriate length and copy the temporary result there
-    itemResult = PROTECT(Rf_allocVector(INTSXP, queryResult.currentIndex));
-    memcpy(INTEGER(itemResult), queryResult.indexList, queryResult.currentIndex * sizeof(int));
+    itemResult = PROTECT(Rf_allocVector(REALSXP, queryResult.currentIndex));
+    memcpy(REAL(itemResult), queryResult.indexList, queryResult.currentIndex * sizeof(double));
 
     // set result to this vector
     SET_VECTOR_ELT(result, i, itemResult);
     UNPROTECT(1); // itemResult
   }
 
-  GEOS_FINISH();
-  UNPROTECT(2); // result, tempItemResult
+    UNPROTECT(1); // result
   return result;
 }
 
@@ -163,8 +173,8 @@ SEXP strtree_query_base(SEXP treeExternalPtr, SEXP geom, GEOSQueryCallback callb
 // the may_intersect querier says yes to everything
 void strtree_callback_may_intersect(void* item, void* userdata) {
   struct QueryResult* queryResult = (struct QueryResult*) userdata;
-  int itemInt = *((int*) item);
-  queryResult->indexList[queryResult->currentIndex] = itemInt;
+  double itemIndex = *((double*) item);
+  queryResult->indexList[queryResult->currentIndex] = itemIndex;
   queryResult->currentIndex++;
 }
 
@@ -178,9 +188,9 @@ SEXP geos_c_strtree_query(SEXP treeExternalPtr, SEXP geom) {
 // will longjmp and we're being called from C++ across a .so boundary
 #define GEOS_STRTREE_CALLBACK(_func)                                           \
   struct QueryResult* queryResult = (struct QueryResult*) userdata;            \
-  int itemInt = *((int*) item);                                                \
+  double itemIndex = *((double*) item);                                        \
                                                                                \
-  SEXP itemGeomPtr = VECTOR_ELT(queryResult->geom, itemInt - 1);               \
+  SEXP itemGeomPtr = VECTOR_ELT(queryResult->geom, itemIndex - 1);             \
   GEOSGeometry* geometryIndex = (GEOSGeometry*) R_ExternalPtrAddr(itemGeomPtr);\
   if (geometryIndex == NULL) {                                                 \
     return;                                                                    \
@@ -193,7 +203,7 @@ SEXP geos_c_strtree_query(SEXP treeExternalPtr, SEXP geom) {
   );                                                                           \
                                                                                \
   if (resultCode) {                                                            \
-    queryResult->indexList[queryResult->currentIndex] = itemInt;               \
+    queryResult->indexList[queryResult->currentIndex] = itemIndex;             \
     queryResult->currentIndex++;                                               \
   }
 
@@ -264,9 +274,9 @@ SEXP geos_c_covered_by_matrix(SEXP geom, SEXP treeExternalPtr) {
 // quiet approach to errors because this is called across a .so boundary
 void strtree_callback_equals(void* item, void* userdata) {
   struct QueryResult* queryResult = (struct QueryResult*) userdata;
-  int itemInt = *((int*) item);
+  double itemIndex = *((double*) item);
 
-  SEXP itemGeomPtr = VECTOR_ELT(queryResult->geom, itemInt - 1);
+  SEXP itemGeomPtr = VECTOR_ELT(queryResult->geom, itemIndex - 1);
   GEOSGeometry* geometryIndex = (GEOSGeometry*) R_ExternalPtrAddr(itemGeomPtr);
   // don't know how this could happen, as the tree was already checked for
   // NULL (which could happen on save/reload)
@@ -281,7 +291,7 @@ void strtree_callback_equals(void* item, void* userdata) {
   );
 
   if (resultCode) {
-    queryResult->indexList[queryResult->currentIndex] = itemInt;
+    queryResult->indexList[queryResult->currentIndex] = itemIndex;
     queryResult->currentIndex++;
   }
 }
@@ -294,9 +304,9 @@ SEXP geos_c_equals_matrix(SEXP geom, SEXP treeExternalPtr) {
 // quiet approach to errors because this is called accross a .so boundary
 void strtree_callback_equals_exact(void* item, void* userdata) {
   struct QueryResult* queryResult = (struct QueryResult*) userdata;
-  int itemInt = *((int*) item);
+  double itemIndex = *((double*) item);
 
-  SEXP itemGeomPtr = VECTOR_ELT(queryResult->geom, itemInt - 1);
+  SEXP itemGeomPtr = VECTOR_ELT(queryResult->geom, itemIndex - 1);
   GEOSGeometry* geometryIndex = (GEOSGeometry*) R_ExternalPtrAddr(itemGeomPtr);
   // don't know how this could happen, as the tree was already checked for
   // NULL (which could happen on save/reload)
@@ -314,7 +324,7 @@ void strtree_callback_equals_exact(void* item, void* userdata) {
   );
 
   if (resultCode) {
-    queryResult->indexList[queryResult->currentIndex] = itemInt;
+    queryResult->indexList[queryResult->currentIndex] = itemIndex;
     queryResult->currentIndex++;
   }
 }
@@ -359,15 +369,17 @@ SEXP geos_strtree_nearest_base(SEXP treeExternalPtr, SEXP geom,
                                GEOSDistanceCallback callback, SEXP extra) {
   GEOSSTRtree* tree = (GEOSSTRtree*) R_ExternalPtrAddr(treeExternalPtr);
 
-  if (tree == NULL && Rf_xlength(R_ExternalPtrTag(treeExternalPtr)) > 0) {
+  if (tree == NULL) {
     UNPROTECT(1);
     Rf_error("External pointer (geos_strtree) is not valid");
   }
 
+  R_xlen_t treeSize = Rf_xlength(geos_strtree_indices(treeExternalPtr));
+
   // allocate the list() result
   R_xlen_t size = Rf_xlength(geom);
-  SEXP result = PROTECT(Rf_allocVector(INTSXP, size));
-  int* pResult = INTEGER(result);
+  SEXP result = PROTECT(Rf_allocVector(REALSXP, size));
+  double* pResult = REAL(result);
 
   GEOS_INIT();
 
@@ -379,38 +391,35 @@ SEXP geos_strtree_nearest_base(SEXP treeExternalPtr, SEXP geom,
 
   SEXP item;
   GEOSGeometry* geometry;
-  int* itemResult;
+  double* itemResult;
   for (R_xlen_t i = 0; i < size; i++) {
     item = VECTOR_ELT(geom, i);
 
     if (item == R_NilValue) {
-      pResult[i] = NA_INTEGER;
+      pResult[i] = NA_REAL;
       continue;
     }
 
     geometry = (GEOSGeometry*) R_ExternalPtrAddr(item);
     GEOS_CHECK_GEOMETRY(geometry, i);
 
-    // tree is only NULL here as an empty tree
-    if (tree != NULL) {
-      itemResult = (int*) GEOSSTRtree_nearest_generic_r(
+    if (treeSize > 0) {
+      itemResult = (double*) GEOSSTRtree_nearest_generic_r(
         handle, tree, geometry, geometry,
         callback, &distanceInfo
       );
 
       if (itemResult == NULL) {
-        UNPROTECT(1); // result
-        GEOS_ERROR("[i=%d] ", i + 1);
+        Rf_error("[%d] %s", i + 1, globalErrorMessage);
       }
 
       pResult[i] = *itemResult;
     } else {
-      pResult[i] = NA_INTEGER;
+      pResult[i] = NA_REAL;
     }
   }
 
-  GEOS_FINISH();
-  UNPROTECT(1); // result
+    UNPROTECT(1); // result
   return result;
 }
 
@@ -428,7 +437,7 @@ SEXP geos_c_nearest_error(SEXP geom, SEXP treeExternalPtr) {
 
 #define GEOS_CALLBACK_DIST(_call)                              \
   struct DistanceQueryInfo* info = (struct DistanceQueryInfo*) userdata;\
-  SEXP treeGeomExternalPtr = VECTOR_ELT(info->treeData, (*((int*) item1)) - 1); \
+  SEXP treeGeomExternalPtr = VECTOR_ELT(info->treeData, (*((double*) item1)) - 1); \
   GEOSGeometry* treeGeometry = (GEOSGeometry*) R_ExternalPtrAddr(treeGeomExternalPtr); \
   GEOSGeometry* itemGeometry = (GEOSGeometry*) item2;            \
   return _call;
